import dataclasses
from enum import Enum, EnumMeta
from typing import List, Any

from workflow.constants import AZURE_EXPORT_ACCOUNT, AZURE_EXPORT_CONTAINER, AZURE_EXPORT_DATA_PATH, \
    AZURE_EXPORT_NUM_FILES, AZURE_EXPORT_SAS, AZURE_IMPORT_ACCOUNT, AZURE_IMPORT_CONTAINER, AZURE_IMPORT_DATA_PATH, \
    AZURE_IMPORT_SAS
from railib import api


class MetaEnum(EnumMeta):
    def __contains__(cls, item):
        try:
            cls(item)
        except ValueError:
            return False
        return True


class BaseEnum(Enum, metaclass=MetaEnum):
    pass


class FileFormat(str, BaseEnum):
    CSV = '.csv'
    JSON = '.json'
    JSONL = '.jsonl'
    CSV_GZ = '.csv.gz'
    JSON_GZ = '.csv.gz'
    JSONL_GZ = '.csv.gz'

    @staticmethod
    def is_supported(file_name):
        """
        Check if file has supported format
        :param file_name: File name
        :return: True if a file name ends with supported file format
        """
        for f in FileFormat:
            if file_name.endswith(f):
                return True
        return False


class FileType(str, BaseEnum):
    CSV = 'CSV'
    JSON = 'JSON'
    JSONL = 'JSONL'


class SourceType(Enum):
    LOCAL = 1
    REMOTE = 2

    @staticmethod
    def from_source(src):
        if 'is_local' in src and src['is_local'] == 'Y':
            return SourceType.LOCAL
        elif 'is_remote' in src and src['is_remote'] == 'Y':
            return SourceType.REMOTE
        else:
            raise ValueError("Source is neither local nor remote.")


@dataclasses.dataclass
class Source:
    relation: str
    input_format: str
    paths: List[str]
    is_partitioned: bool
    is_master: bool

    def to_paths_csv(self) -> str:
        return "\n".join([f"{self.relation},{path}" for path in self.paths])

    def to_formats_csv(self) -> str:
        return f"{self.relation},{self.input_format.upper()}"


@dataclasses.dataclass
class RaiConfig:
    ctx: api.Context
    engine: str
    database: str


@dataclasses.dataclass
class EnvConfig:
    # Azure EXPORT blob
    azure_export_account: str
    azure_export_container: str
    azure_export_data_path: str
    azure_export_num_files: int
    azure_export_sas: str
    # Azure IMPORT blob
    azure_import_account: str
    azure_import_container: str
    azure_import_data_path: str
    azure_import_sas: str

    @staticmethod
    def from_env_vars(env_vars: dict[str, Any]):
        azure_export_account = env_vars[AZURE_EXPORT_ACCOUNT] if AZURE_EXPORT_ACCOUNT in env_vars else ""
        azure_export_container = env_vars[
            AZURE_EXPORT_CONTAINER] if AZURE_EXPORT_CONTAINER in env_vars else ""
        azure_export_data_path = env_vars[
            AZURE_EXPORT_DATA_PATH] if AZURE_EXPORT_DATA_PATH in env_vars else ""
        azure_export_num_files = env_vars[
            AZURE_EXPORT_NUM_FILES] if AZURE_EXPORT_NUM_FILES in env_vars else 1
        azure_export_sas = env_vars[AZURE_EXPORT_SAS] if AZURE_EXPORT_SAS in env_vars else ""
        azure_import_account = env_vars[AZURE_IMPORT_ACCOUNT] if AZURE_IMPORT_ACCOUNT in env_vars else ""
        azure_import_container = env_vars[
            AZURE_IMPORT_CONTAINER] if AZURE_IMPORT_CONTAINER in env_vars else ""
        azure_import_data_path = env_vars[
            AZURE_IMPORT_DATA_PATH] if AZURE_IMPORT_DATA_PATH in env_vars else ""
        azure_import_sas = env_vars[AZURE_IMPORT_SAS] if AZURE_IMPORT_SAS in env_vars else ""
        return EnvConfig(
            azure_export_account=azure_export_account,
            azure_export_container=azure_export_container,
            azure_export_data_path=azure_export_data_path,
            azure_export_num_files=azure_export_num_files,
            azure_export_sas=azure_export_sas,
            azure_import_account=azure_import_account,
            azure_import_container=azure_import_container,
            azure_import_data_path=azure_import_data_path,
            azure_import_sas=azure_import_sas
        )


@dataclasses.dataclass
class Export:
    meta_key: List[str]
    relation: str
    relative_path: str
    file_type: FileType
    offset_by_number_of_days: int = 0


@dataclasses.dataclass
class BatchConfig:
    name: str
    content: str



import logging
from typing import List
from azure.storage.blob import BlobServiceClient

from workflow.common import FileFormat
from workflow.common import EnvConfig
from workflow.constants import BLOB_PAGE_SIZE


def list_files_in_containers(logger: logging.Logger, env_config: EnvConfig, path_prefix) -> List[str]:
    azure_account = env_config.azure_import_account
    azure_container = env_config.azure_import_container
    account_url = f"https://{azure_account}.blob.core.windows.net"

    blob_service_client = BlobServiceClient(account_url=account_url, credential=env_config.azure_import_sas)
    container_client = blob_service_client.get_container_client(azure_container)

    # Get a list of blobs in the folder
    logger.debug(f"Path prefix to list blob files: {path_prefix}")
    paths = []
    for page in container_client.list_blobs(name_starts_with=path_prefix, results_per_page=BLOB_PAGE_SIZE).by_page():
        for blob in page:
            blob_name = blob.name
            if FileFormat.is_supported(blob_name):
                paths.append(f"azure://{azure_account}.blob.core.windows.net/{azure_container}/{blob_name}")
            else:
                logger.debug(f"Skip unsupported file from blob: {blob_name}")
    return paths



import logging
import sys
import random
import dataclasses
from typing import List, Iterable

from workflow import utils
from workflow.common import EnvConfig, FileType, Export, Source, SourceType
from workflow.constants import IMPORT_CONFIG_REL, FILE_LOAD_RELATION

# Static queries
DISABLE_IVM = "def insert:relconfig:disable_ivm = true"


@dataclasses.dataclass
class QueryWithInputs:
    query: str
    inputs: dict


def load_json(relation: str, data) -> QueryWithInputs:
    return QueryWithInputs(f"def config:data = data\n" f"def insert:{relation} = load_json[config]", {"data": data})


def install_model(models: dict) -> QueryWithInputs:
    queries_inputs = {}
    rand_uint = random.randint(0, sys.maxsize)

    index = 0
    query = ""
    for name in models:
        input_name = f"input_{str(rand_uint)}_{index}"
        query += f"def delete:rel:catalog:model[\"{name}\"] = rel:catalog:model[\"{name}\"]\n" \
                 f"def insert:rel:catalog:model[\"{name}\"] = {input_name}\n"
        queries_inputs[input_name] = models[name]
        index += 1

    return QueryWithInputs(query, queries_inputs)


def populate_source_configs(sources: List[Source]) -> str:
    source_config_csv = "\n".join([source.to_paths_csv() for source in sources])
    data_formats_csv = "\n".join([source.to_formats_csv() for source in sources])

    simple_sources = list(filter(lambda source: not source.is_partitioned, sources))
    multipart_sources = list(filter(lambda source: source.is_partitioned, sources))
    master_sources = list(filter(lambda source: source.is_master, sources))

    return f"""
        def resource_config[:data] = \"\"\"{source_config_csv}\"\"\"
        def resource_config[:syntax, :header_row] = -1
        def resource_config[:syntax, :header] = (1, :Relation); (2, :Path)
        def resource_config[:schema, :Relation] = "string"
        def resource_config[:schema, :Path] = "string"

        def source_config_csv = load_csv[resource_config]

        def insert:source_declares_resource(r, p) =
            exists(i : source_config_csv(:Relation, i, r) and source_config_csv(:Path, i, p))

        def input_format_config[:data] = \"\"\"{data_formats_csv}\"\"\"

        def input_format_config[:syntax, :header_row] = -1
        def input_format_config[:syntax, :header] = (1, :Relation); (2, :InputFormatCode)
        def input_format_config[:schema, :Relation] = "string"
        def input_format_config[:schema, :InputFormatCode] = "string"

        def input_format_config_csv = load_csv[input_format_config]

        def insert:source_has_input_format(r,p) =
            exists(i : input_format_config_csv(:Relation, i, r) and input_format_config_csv(:InputFormatCode, i, p))

        {f"def insert:simple_relation = {_to_rel_literal_relation([source.relation for source in simple_sources])}" if len(simple_sources) > 0 else ""}
        {f"def insert:multi_part_relation = {_to_rel_literal_relation([source.relation for source in multipart_sources])}" if len(multipart_sources) > 0 else ""}
        {f"def insert:master_source = {_to_rel_literal_relation([source.relation for source in master_sources])}" if len(master_sources) > 0 else ""}
    """


def load_resources(logger: logging.Logger, env_config: EnvConfig, resources, src) -> str:
    rel_name = src["source"]

    file_stype_str = src["file_type"]
    file_type = FileType[file_stype_str]
    src_type = SourceType.from_source(src)

    if 'is_multi_part' in src and src['is_multi_part'] == 'Y':
        if file_type == FileType.CSV or file_type == FileType.JSONL:
            if src_type == SourceType.LOCAL:
                logger.info(f"Loading {len(resources)} shards from local files")
                return _local_load_multipart_query(rel_name, file_type, resources)
            elif src_type == SourceType.REMOTE:
                logger.info(f"Loading {len(resources)} shards from remote files")
                return _remote_load_multipart_query(rel_name, file_type, resources, env_config)
        else:
            logger.error(f"Unknown file type {file_stype_str}")
    else:
        if src_type == SourceType.LOCAL:
            logger.info("Loading from local file.")
            return _local_load_simple_query(rel_name, resources[0]["uri"], file_type)
        elif src_type == SourceType.REMOTE:
            logger.info("Loading from remote file.")
            return _remote_load_simple_query(rel_name, resources[0]["uri"], file_type, env_config)


def export_relations_local(logger: logging.Logger, exports: List[Export]) -> str:
    query = ""
    for export in exports:
        if export.file_type == FileType.CSV:
            if export.meta_key:
                query += _export_meta_relation_as_csv_local(export)
            else:
                query += _export_relation_as_csv_local(export.relation)
        else:
            logger.warning(f"Unsupported export type: {export.file_type}")
    return query


def export_relations_remote(logger: logging.Logger, env_config: EnvConfig, exports: List[Export], end_date: str,
                            date_format: str) -> str:
    query = f"""
    def _credentials_config:integration:provider = "azure"
    def _credentials_config:integration:credentials:azure_sas_token = raw"{env_config.azure_export_sas}"
    """
    for export in exports:
        if export.file_type == FileType.CSV:
            if export.meta_key:
                query += _export_meta_relation_as_csv_remote(env_config, export, end_date, date_format)
            else:
                query += _export_relation_as_csv_remote(env_config, export, end_date, date_format)
        else:
            logger.warning(f"Unsupported export type: {export.file_type}")
    return query


def init_workflow_steps(batch_config_name: str) -> str:
    return f"""
    def delete:batch_workflow_step:state_value(s, v) {{ 
        batch_workflow_step:workflow[s] . batch_workflow:name[:{batch_config_name}] and
        batch_workflow_step:state_value(s, v)
    }}
    def delete:batch_workflow_step:execution_time_value(s, v) {{ 
        batch_workflow_step:workflow[s] . batch_workflow:name[:{batch_config_name}] and
        batch_workflow_step:execution_time_value(s, v)
    }}
    def insert:batch_workflow_step:execution_time_value(s, v) {{ 
        batch_workflow_step:workflow[s] . batch_workflow:name[:{batch_config_name}] and
        v = 0.0
        
    }}
    def insert:batch_workflow_step:state_value(s, v) {{ 
        batch_workflow_step:workflow[s] . batch_workflow:name[:{batch_config_name}] and
        v = "INIT"
    }}
    """


def update_step_state(idt: str, state: str) -> str:
    return f"""
    def insert:batch_workflow_step:state_value(s in BatchWorkflowStep, v) {{
        s = uint128_hash_value_convert[parse_uuid["{idt}"]] and 
        v = "{state}"
    }}
    """


def update_execution_time(idt: str, execution_time: float) -> str:
    return f"""
    def insert:batch_workflow_step:execution_time_value(s in BatchWorkflowStep, v) {{
        s = uint128_hash_value_convert[parse_uuid["{idt}"]] and 
        v = {execution_time}
    }}
    """


def materialize(relations: List[str]) -> str:
    query = ""
    for relation in relations:
        query += f"def output:{relation} = count[{relation}]\n"
    return query


def output_json(relation: str) -> str:
    return f"def output = json_string[{relation}]"


def _local_load_simple_query(rel_name: str, uri: str, file_type: FileType) -> str:
    try:
        raw_data_rel_name = f"{rel_name}_data"
        data = utils.read(uri)
        return f"{_load_from_literal(data, raw_data_rel_name)}\n" \
               f"def {IMPORT_CONFIG_REL}:{rel_name}:data = {raw_data_rel_name}\n" \
               f"{_simple_insert_query(rel_name, file_type)}\n"
    except OSError as e:
        raise e


def _remote_load_simple_query(rel_name: str, uri: str, file_type: FileType, env_config: EnvConfig) -> str:
    return f"def {IMPORT_CONFIG_REL}:{rel_name}:integration:provider = \"azure\"\n" \
           f"def {IMPORT_CONFIG_REL}:{rel_name}:integration:credentials:azure_sas_token = raw\"{env_config.azure_import_sas}\"\n" \
           f"def {IMPORT_CONFIG_REL}:{rel_name}:path = \"{uri}\"\n" \
           f"{_simple_insert_query(rel_name, file_type)}"


def _local_load_multipart_query(rel_name: str, file_type: FileType, parts) -> str:
    raw_data_rel_name = f"{rel_name}_data"

    raw_text = ""
    part_indexes = ""
    for part in parts:
        try:
            part_idx = part["part_index"]
            data = utils.read(part["uri"])
            raw_text += _load_from_indexed_literal(data, raw_data_rel_name, part_idx)
            part_indexes += f"{part_idx}\n"
        except OSError as e:
            raise e

    insert_text = _multi_part_insert_query(rel_name, file_type)
    load_config = _multi_part_load_config_query(rel_name, file_type,
                                                _local_multipart_config_integration(raw_data_rel_name))

    return f"{_part_index_relation(part_indexes)}\n" \
           f"{raw_text}\n" \
           f"{load_config}\n" \
           f"{insert_text}"


def _remote_load_multipart_query(rel_name: str, file_type: FileType, parts, env_config: EnvConfig) -> str:
    path_rel_name = f"{rel_name}_path"

    part_indexes = ""
    part_uri_map = ""
    for part in parts:
        part_idx = part["part_index"]
        part_uri = part["uri"]
        part_indexes += f"{part_idx}\n"
        part_uri_map += f"{part_idx},\"{part_uri}\"\n"

    insert_text = _multi_part_insert_query(rel_name, file_type)
    load_config = _multi_part_load_config_query(rel_name, file_type,
                                                _remote_multipart_config_integration(path_rel_name,
                                                                                     env_config))

    return f"{_part_index_relation(part_indexes)}\n" \
           f"{_path_rel_name_relation(path_rel_name, part_uri_map)}\n" \
           f"{load_config}\n" \
           f"{insert_text}"


def _multi_part_load_config_query(rel_name: str, file_type: FileType, config_integration: str) -> str:
    schema = ""
    if file_type == FileType.CSV:
        schema = f"def schema = {IMPORT_CONFIG_REL}:{rel_name}:schema\n" \
                 f"def syntax:header = {IMPORT_CONFIG_REL}:{rel_name}:syntax:header"

    return f"""
module {_config_rel_name(rel_name)}[i in part_indexes]
    {schema}
    {config_integration}
end
"""


def _local_multipart_config_integration(raw_data_rel_name: str) -> str:
    return f"def data = {raw_data_rel_name}[i]"


def _remote_multipart_config_integration(path_rel_name: str, env_config: EnvConfig) -> str:
    return f"def integration:provider = \"azure\"\n" \
           f"def integration:credentials:azure_sas_token = raw\"{env_config.azure_import_sas}\"\n" \
           f"def path = {path_rel_name}[i]\n"


def _multi_part_insert_query(rel_name: str, file_type: FileType) -> str:
    conf_rel_name = _config_rel_name(rel_name)

    return f"def insert:source_catalog:{rel_name}[i] = {FILE_LOAD_RELATION[file_type]}[{conf_rel_name}[i]]"


def _simple_insert_query(rel_name: str, file_type: FileType) -> str:
    return f"def insert:simple_source_catalog:{rel_name} = {FILE_LOAD_RELATION[file_type]}[{IMPORT_CONFIG_REL}:{rel_name}]"


def _load_from_indexed_literal(data: str, raw_data_rel_name: str, index: int) -> str:
    return f"def {raw_data_rel_name}[{index}] =\n" \
           f"raw\"\"\"{data}" \
           f"\"\"\""


def _load_from_literal(data: str, raw_data_rel_name: str) -> str:
    return f"def {raw_data_rel_name} =\n" \
           f"raw\"\"\"{data}" \
           f"\"\"\""


def _config_rel_name(rel: str) -> str: return f"load_{rel}_config"


def _part_index_relation(part_index: str) -> str:
    return f"def part_index_config:schema:INDEX = \"int\"\n" \
           f"def part_index_config:data = \"\"\"\n" \
           f"INDEX\n" \
           f"{part_index}\n" \
           f"\"\"\"" \
           f"def part_indexes_csv = load_csv[part_index_config]\n" \
           f"def part_indexes = part_indexes_csv:INDEX[_]"


def _path_rel_name_relation(path_rel_name: str, part_uri_map: str) -> str:
    return f"def part_uri_map_config:schema:INDEX = \"int\"\n" \
           f"def part_uri_map_config:schema:URI = \"string\"\n" \
           f"def part_uri_map_config:data = \"\"\"\n" \
           f"INDEX,URI\n" \
           f"{part_uri_map}\n" \
           f"\"\"\"" \
           f"def part_uri_map_csv = load_csv[part_uri_map_config]\n" \
           f"def {path_rel_name}(i, u) {{ part_uri_map_csv:INDEX(row, i) and part_uri_map_csv:URI(row, u) from row }}"


def _export_relation_as_csv_local(rel_name) -> str:
    return f"def _export_csv_config:{rel_name} = export_config:{rel_name}\n" \
           f"def output:{rel_name} = csv_string[_export_csv_config:{rel_name}]"


def _export_meta_relation_as_csv_local(export: Export) -> str:
    rel_name = export.relation
    key_str = _to_rel_meta_key_as_seq(export)
    return f"""
    module _export_csv_config
        def {rel_name}[{key_str}] =
            export_config:{rel_name}[{key_str}], export_config:{rel_name}:meta_key({key_str})
    end
    def output:{rel_name}[{key_str}] = csv_string[_export_csv_config:{rel_name}[{key_str}]]
    """


def _export_relation_as_csv_remote(env_config: EnvConfig, export: Export, end_date: str, date_format: str) -> str:
    rel_name = export.relation
    export_path = f"{_compose_export_path(env_config, export, end_date, date_format)}/{rel_name}.csv"
    return f"""
    module _export_csv_config
        def {rel_name} = export_config:{rel_name}
        def {rel_name}:path = raw"{export_path}"
        def {rel_name} = _credentials_config
    end
    def export:{rel_name} = export_csv[_export_csv_config:{rel_name}]
    """


def _export_meta_relation_as_csv_remote(env_config: EnvConfig, export: Export, end_date: str, date_format: str) -> str:
    rel_name = export.relation
    postfix = _to_rel_meta_key_as_str(export)
    base_path = _compose_export_path(env_config, export, end_date, date_format)
    export_path = f"{base_path}/{rel_name}_{postfix}.csv"
    key_str = _to_rel_meta_key_as_seq(export)
    return f"""
    module _export_csv_config
        module {rel_name}
            def meta_key({key_str}) = export_config:{rel_name}:meta_key({key_str})
            def filename_postfix[{key_str}] = meta_key({key_str}), "{_to_rel_meta_key_as_str(export)}"
            def path[{key_str}] = meta_key({key_str}), "{export_path}"
    
            def config[keys...] = meta_key(keys...), {{
                :path, path[keys...] ;
                export_config:{rel_name}[keys...] ;
                _credentials_config
            }}
        end
    end
    def export:{rel_name}[{key_str}] = export_csv[_export_csv_config:{rel_name}:config[{key_str}]],
        _export_csv_config:{rel_name}:meta_key[{key_str}]
    """


def _compose_export_path(env_config: EnvConfig, export: Export, end_date: str, date_format: str) -> str:
    account_url = f"azure://{env_config.azure_export_account}.blob.core.windows.net"
    container_path = env_config.azure_export_container
    folder_path = env_config.azure_export_data_path
    date_path = utils.get_date_path(end_date, date_format, export.offset_by_number_of_days)
    if export.meta_key:
        postfix = _to_rel_meta_key_as_str(export)
        return f"{account_url}/{container_path}/{folder_path}/{export.relative_path}_{postfix}/{date_path}"
    else:
        return f"{account_url}/{container_path}/{folder_path}/{export.relative_path}/{date_path}"


def _to_rel_literal_relation(xs: Iterable[str]) -> str:
    return "{ " + " ; ".join([f"\"{x}\"" for x in xs]) + " }"


def _to_rel_meta_key_as_seq(export: Export) -> str:
    key_values = range(0, len(export.meta_key))
    return ", ".join([f"_v{i}" for i in key_values])


def _to_rel_meta_key_as_str(export: Export) -> str:
    key_values = range(0, len(export.meta_key))
    return "_".join([f"%(_v{i})" for i in key_values])



import logging
import time
import dataclasses
from more_itertools import peekable
from typing import List, Dict
from enum import Enum

from workflow.utils import save_csv_output, format_duration, build_models, extract_date_range, build_relation_path, \
    get_common_model_relative_path
from workflow.common import EnvConfig, RaiConfig, Source, BatchConfig, Export, FileType
from workflow.manager import ResourceManager
from workflow import query as q, paths, rai, constants
from types import MappingProxyType


CONFIGURE_SOURCES = 'ConfigureSources'
INSTALL_MODELS = 'InstallModels'
LOAD_DATA = 'LoadData'
# not supported
INVOKE_SOLVER = 'InvokeSolver'
MATERIALIZE = 'Materialize'
EXPORT = 'Export'


class WorkflowRunMode(Enum):
    LOCAL = 'local'
    REMOTE = 'remote'

    def __str__(self):
        return self.value


class WorkflowStepState(str, Enum):
    INIT = 'INIT'
    IN_PROGRESS = 'IN_PROGRESS'
    SUCCESS = 'SUCCESS'
    FAILED = 'FAILED'


@dataclasses.dataclass
class WorkflowConfig:
    env: EnvConfig
    run_mode: WorkflowRunMode
    batch_config: BatchConfig
    recover: bool
    recover_step: str
    selected_steps: List[str]
    step_params: dict


class WorkflowStep:
    idt: str
    name: str
    state: WorkflowStepState
    timing: int
    engine_size: str

    def __init__(self, idt, name, state, timing, engine_size):
        self.idt = idt
        self.name = name
        self.state = state
        self.time = timing
        self.engine_size = engine_size

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        raise NotImplementedError("This class is abstract")


class WorkflowStepFactory:

    def get_step(self, logger: logging.Logger, config: WorkflowConfig, step: dict) -> WorkflowStep:
        idt = step["idt"]
        name = step["name"]
        engine_size = str.upper(step["engineSize"]) if "engineSize" in step else None
        state = WorkflowStepState(step.get("state"))
        timing = step.get("executionTime", 0)

        self._validate_params(config, step)
        return self._get_step(logger, config, idt, name, state, timing, engine_size, step)

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        raise NotImplementedError("This class is abstract")

    def _required_params(self, config: WorkflowConfig) -> List[str]:
        return []

    def _validate_params(self, config: WorkflowConfig, step: dict) -> None:
        params = self._required_params(config)
        missed_params = []
        for param in params:
            if param not in config.step_params:
                missed_params.append(param)
        if missed_params:
            raise ValueError(f"Workflow config doesn't have required parameters for {step['name']} : '{missed_params}'")


class InstallModelsStep(WorkflowStep):
    rel_config_dir: str
    model_files: List[str]

    def __init__(self, idt, name, state, timing, engine_size, rel_config_dir, model_files):
        super().__init__(idt, name, state, timing, engine_size)
        self.rel_config_dir = rel_config_dir
        self.model_files = model_files

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        logger.info("Executing InstallModel step..")

        rai.install_models(logger, rai_config, build_models(self.model_files, self.rel_config_dir))


class InstallModelWorkflowStepFactory(WorkflowStepFactory):

    def _required_params(self, config: WorkflowConfig) -> List[str]:
        return [constants.REL_CONFIG_DIR]

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        rel_config_dir = config.step_params[constants.REL_CONFIG_DIR]
        return InstallModelsStep(idt, name, state, timing, engine_size, rel_config_dir, step["modelFiles"])


class ConfigureSourcesWorkflowStep(WorkflowStep):
    config_files: List[str]
    rel_config_dir: str
    sources: List[Source]

    def __init__(self, idt, name, state, timing, engine_size, config_files, rel_config_dir, sources):
        super().__init__(idt, name, state, timing, engine_size)
        self.config_files = config_files
        self.rel_config_dir = rel_config_dir
        self.sources = sources

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        logger.info("Executing ConfigureSources step..")

        rai.install_models(logger, rai_config, build_models(self.config_files, self.rel_config_dir))

        rai.execute_query(logger, rai_config, q.populate_source_configs(self.sources), readonly=False)


class ConfigureSourcesWorkflowStepFactory(WorkflowStepFactory):

    def _required_params(self, config: WorkflowConfig) -> List[str]:
        required_params = [constants.REL_CONFIG_DIR, constants.START_DATE, constants.END_DATE]
        if config.run_mode == WorkflowRunMode.LOCAL:
            required_params.append(constants.LOCAL_DATA_DIR)
        return required_params

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        if config.run_mode == WorkflowRunMode.LOCAL:
            local_data_dir = config.step_params[constants.LOCAL_DATA_DIR]
            paths_builder = paths.LocalPathsBuilder(local_data_dir)
        elif config.run_mode == WorkflowRunMode.REMOTE:
            paths_builder = paths.RemotePathsBuilder(config.env)
        else:
            raise Exception("unsupported mode")
        rel_config_dir = config.step_params[constants.REL_CONFIG_DIR]
        start_date = config.step_params[constants.START_DATE]
        end_date = config.step_params[constants.END_DATE]
        return ConfigureSourcesWorkflowStep(idt, name, state, timing, engine_size, step["configFiles"], rel_config_dir,
                                            self._inflate_sources(logger, step["sources"], start_date, end_date,
                                                                  paths_builder))

    @staticmethod
    def _inflate_sources(logger: logging.Logger, sources: List[Dict], start_date, end_date,
                         paths_builder: paths.PathsBuilder) -> List[Source]:
        result = []
        for source in sources:
            if "future" not in source or not source["future"]:
                relation = source["relation"]
                relative_path = source["relativePath"]
                input_format = source["inputFormat"]
                extensions = source.get("extensions", [input_format])
                is_partitioned = source["isPartitioned"]
                is_master = source.get("isMaster", False)
                loads_number_of_days = source.get("loadsNumberOfDays")
                offset_by_number_of_days = source.get("offsetByNumberOfDays")
                result.append(Source(
                    relation,
                    input_format,
                    paths_builder.build(logger,
                                        extract_date_range(logger, start_date, end_date, loads_number_of_days,
                                                           offset_by_number_of_days),
                                        relative_path,
                                        extensions,
                                        is_master),
                    is_partitioned,
                    is_master
                ))
        return result


class LoadDataWorkflowStep(WorkflowStep):
    collapse_partitions_on_load: bool

    def __init__(self, idt, name, state, timing, engine_size, collapse_partitions_on_load):
        super().__init__(idt, name, state, timing, engine_size,)
        self.collapse_partitions_on_load = collapse_partitions_on_load

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        logger.info("Executing LoadData step..")

        missed_resources = rai.execute_relation_json(logger, rai_config, constants.MISSED_RESOURCES_REL)

        if not missed_resources:
            logger.info("Missed resources list is empty")

        for src in missed_resources:
            self._load_source(logger, env_config, rai_config, src)

    def _load_source(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig, src):
        source_name = src["source"]
        if 'is_master' in src and src['is_master'] == 'F':
            if self.collapse_partitions_on_load:
                srcs = src["dates"]
                first_date = srcs[0]["date"]
                last_date = srcs[-1]["date"]

                logger.info(
                    f"Loading '{source_name}' from all partitions simultaneously, range {first_date} to {last_date}")

                resources = []
                for d in srcs:
                    resources += d["resources"]
                self._load_resource(logger, env_config, rai_config, resources, src)
            else:
                logger.info(f"Loading '{source_name}' one partition at a time")
                for d in src["dates"]:
                    logger.info(f"Loading partition for date {d['date']}")

                    resources = d["resources"]
                    self._load_resource(logger, env_config, rai_config, resources, src)
        else:
            logger.info(f"Loading master source'{source_name}'")
            self._load_resource(logger, env_config, rai_config, src["resources"], src)

    def _load_resource(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig, resources, src) -> None:
        try:
            rai.execute_query(logger, rai_config, q.load_resources(logger, env_config, resources, src), readonly=False)
        except KeyError as e:
            logger.error(f"Unsupported file type: {src['file_type']}. Skip the source: {src}", e)
        except ValueError as e:
            logger.error(f"Unsupported source type. Skip the source: {src}", e)


class LoadDataWorkflowStepFactory(WorkflowStepFactory):

    def _required_params(self, config: WorkflowConfig) -> List[str]:
        return [constants.COLLAPSE_PARTITIONS_ON_LOAD]

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        collapse_partitions_on_load = config.step_params[constants.COLLAPSE_PARTITIONS_ON_LOAD]
        return LoadDataWorkflowStep(idt, name, state, timing, engine_size, collapse_partitions_on_load)


class MaterializeWorkflowStep(WorkflowStep):
    relations: List[str]
    materialize_jointly: bool

    def __init__(self, idt, name, state, timing, engine_size, relations, materialize_jointly):
        super().__init__(idt, name, state, timing, engine_size)
        self.relations = relations
        self.materialize_jointly = materialize_jointly

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        logger.info("Executing Materialize step..")

        if self.materialize_jointly:
            rai.execute_query(logger, rai_config, q.materialize(self.relations), readonly=False)
        else:
            for relation in self.relations:
                rai.execute_query(logger, rai_config, q.materialize([relation]), readonly=False)


class MaterializeWorkflowStepFactory(WorkflowStepFactory):

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        return MaterializeWorkflowStep(idt, name, state, timing, engine_size, step["relations"],
                                       step["materializeJointly"])


class ExportWorkflowStep(WorkflowStep):
    exports: List[Export]
    export_jointly: bool

    def __init__(self, idt, name, state, timing, engine_size, exports, export_jointly):
        super().__init__(idt, name, state, timing, engine_size)
        self.exports = exports
        self.export_jointly = export_jointly

    def execute(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig):
        logger.info("Executing Export step..")

        if self.export_jointly:
            self._export(logger, env_config, rai_config, self.exports)
        else:
            for export in self.exports:
                self._export(logger, env_config, rai_config, [export])

    def _export(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig, exports):
        raise NotImplementedError("This class is abstract")


class ExportWorkflowStepFactory(WorkflowStepFactory):

    def _required_params(self, config: WorkflowConfig) -> List[str]:
        required_params = []
        if config.run_mode == WorkflowRunMode.LOCAL:
            required_params.append(constants.OUTPUT_ROOT)
        elif config.run_mode == WorkflowRunMode.REMOTE:
            required_params.append(constants.END_DATE)
        return required_params

    def _get_step(self, logger: logging.Logger, config: WorkflowConfig, idt, name, state, timing, engine_size,
                  step: dict) -> WorkflowStep:
        exports = self._load_exports(logger, step)
        if config.run_mode == WorkflowRunMode.LOCAL:
            output_root = config.step_params[constants.OUTPUT_ROOT]
            return LocalExportWorkflowStep(idt, name, state, timing, engine_size, exports, step["exportJointly"],
                                           output_root)
        elif config.run_mode == WorkflowRunMode.REMOTE:
            end_date = config.step_params[constants.END_DATE]
            return RemoteExportWorkflowStep(idt, name, state, timing, engine_size, exports, step["exportJointly"],
                                            step["dateFormat"], end_date)
        else:
            raise Exception("Unsupported mode")

    @staticmethod
    def _load_exports(logger: logging.Logger, src) -> List[Export]:
        exports_json = src["exports"]
        exports = []
        for e in exports_json:
            if "future" not in e or not e["future"]:
                try:
                    meta_key = e["metaKey"] if "metaKey" in e else []
                    file_type_str = e["type"].upper()
                    offset_by_number_of_days = e["offsetByNumberOfDays"] if "offsetByNumberOfDays" in e else 0
                    cfg = Export(meta_key, e["configRelName"], e["relativePath"], FileType[file_type_str],
                                 offset_by_number_of_days)
                    exports.append(cfg)
                except KeyError as ex:
                    logger.warning(f"Unsupported FileType: {ex}. Skipping export: {e}.")
        return exports


class LocalExportWorkflowStep(ExportWorkflowStep):
    output_root: str

    def __init__(self, idt, name, state, timing, engine_size, exports, export_jointly, output_root):
        super().__init__(idt, name, state, timing, engine_size, exports, export_jointly)
        self.output_root = output_root

    def _export(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig, exports):
        save_csv_output(
            rai.execute_query_csv(logger, rai_config, q.export_relations_local(logger, exports)),
            self.output_root)


class RemoteExportWorkflowStep(ExportWorkflowStep):
    date_format: str
    end_date: str

    def __init__(self, idt, name, state, timing, engine_size, exports, export_jointly, date_format, end_date):
        super().__init__(idt, name, state, timing, engine_size, exports, export_jointly)
        self.end_date = end_date
        self.date_format = date_format

    def _export(self, logger: logging.Logger, env_config: EnvConfig, rai_config: RaiConfig, exports):
        rai.execute_query(
            logger, rai_config,
            q.export_relations_remote(logger, env_config, exports, self.end_date, self.date_format))


DEFAULT_FACTORIES = MappingProxyType(
    {
        CONFIGURE_SOURCES: ConfigureSourcesWorkflowStepFactory(),
        INSTALL_MODELS: InstallModelWorkflowStepFactory(),
        LOAD_DATA: LoadDataWorkflowStepFactory(),
        MATERIALIZE: MaterializeWorkflowStepFactory(),
        EXPORT: ExportWorkflowStepFactory()
    }
)


class WorkflowExecutor:
    logger: logging.Logger
    config: WorkflowConfig
    resource_manager: ResourceManager
    steps: List[WorkflowStep]

    def __init__(self, logger: logging.Logger, config: WorkflowConfig, resource_manager: ResourceManager, steps):
        self.logger = logger
        self.config = config
        self.resource_manager = resource_manager
        self.steps = steps

    def run(self):
        recover_step_reached = False
        rai_config = self.resource_manager.get_rai_config()
        steps_iter = peekable(self.steps)
        for step in steps_iter:
            if self.config.selected_steps:
                if step.name not in self.config.selected_steps:
                    self.logger.info(f"Step {step.name} (id='{step.idt}') is not selected. Skipping..")
                    continue
            else:
                # `recover_step` option has higher priority than `recover` option
                if self.config.recover_step and not recover_step_reached:
                    if step.name == self.config.recover_step:
                        recover_step_reached = True
                    if not recover_step_reached:
                        self.logger.info(
                            f"Recovery... Skipping the step {step.name} (id='{step.idt}') till reach recovery step")
                        continue
                elif self.config.recover and step.state == WorkflowStepState.SUCCESS:
                    self.logger.info(f"Recovery... Skipping the successful step {step.name} (id='{step.idt}')")
                    continue

            start_time = time.time()
            rai.execute_query(self.logger, rai_config,
                              q.update_step_state(step.idt, WorkflowStepState.IN_PROGRESS.name), readonly=False,
                              ignore_problems=True)
            try:
                if step.engine_size:
                    self.resource_manager.add_engine(step.engine_size)
                    step.execute(self.logger, self.config.env, self.resource_manager.get_rai_config(step.engine_size))
                    next_step = steps_iter.peek(None)
                    if next_step and next_step.engine_size != step.engine_size:
                        self.resource_manager.remove_engine(step.engine_size)
                else:
                    step.execute(self.logger, self.config.env, rai_config)

                end_time = time.time()
                execution_time = end_time - start_time
                query = "\n".join([q.update_step_state(step.idt, WorkflowStepState.SUCCESS.name),
                                   q.update_execution_time(step.idt, execution_time)])
                rai.execute_query(self.logger, rai_config, query, readonly=False)
            except Exception as e:
                rai.execute_query(self.logger, rai_config,
                                  q.update_step_state(step.idt, WorkflowStepState.FAILED.name), readonly=False,
                                  ignore_problems=True)
                if step.engine_size:
                    self.resource_manager.remove_engine(step.engine_size)
                raise e

            self.logger.info(f"{step.name} (id='{step.idt}') finished in {format_duration(execution_time)}")

    def print_timings(self) -> None:
        rai_config = self.resource_manager.get_rai_config()
        relation = build_relation_path(constants.WORKFLOW_JSON_REL, self.config.batch_config.name)
        workflow_info = rai.execute_relation_json(self.logger, rai_config, relation, ignore_problems=True)
        steps = workflow_info["steps"]
        for step in steps:
            execution_time = step["executionTime"]
            self.logger.info(f"{step['name']} (id={step['idt']}) finished in {format_duration(execution_time)}")
        self.logger.info(f"Total workflow execution time is {format_duration(workflow_info['totalTime'])}")

    @staticmethod
    def init(logger: logging.Logger, config: WorkflowConfig, resource_manager: ResourceManager,
             factories: dict[str, WorkflowStepFactory] = MappingProxyType({}),
             models: dict[str, str] = MappingProxyType({})):
        logger = logger.getChild("workflow")
        rai_config = resource_manager.get_rai_config()

        if not config.recover and not config.recover_step:
            # Install common model for workflow manager
            core_models = build_models(constants.COMMON_MODEL, get_common_model_relative_path(__file__))
            extended_models = {**core_models, **models}
            rai.install_models(logger, rai_config, extended_models)
            # Load batch config
            rai.load_json(logger, rai_config,
                          build_relation_path(constants.CONFIG_BASE_RELATION, config.batch_config.name),
                          config.batch_config.content)
            # Init workflow steps
            rai.execute_query(logger, rai_config, q.init_workflow_steps(config.batch_config.name), readonly=False)

        extended_factories = {**DEFAULT_FACTORIES, **factories}
        workflow_info = rai.execute_relation_json(logger, rai_config,
                                                  build_relation_path(constants.WORKFLOW_JSON_REL,
                                                                      config.batch_config.name), ignore_problems=True)
        steps_json = workflow_info["steps"]
        if not steps_json:
            raise ValueError(f"Config `{config.batch_config.name}` doesn't have workflow steps")
        steps = []
        for step in steps_json:
            step_type = step["type"]
            if step_type not in extended_factories:
                logger.warning(f"Step '{step_type}' is not supported")
            else:
                steps.append(extended_factories.get(step_type).get_step(logger, config, step))
        return WorkflowExecutor(logger, config, resource_manager, steps)




COMMON_MODEL = [
    "source_configs/config.rel",
    "batch_config/batch_config.rel",
    "batch_config/workflow/workflow.rel",
    "batch_config/workflow/steps/configure_sources.rel",
    "batch_config/workflow/steps/export.rel",
    "batch_config/workflow/steps/install_models.rel",
    "batch_config/workflow/steps/invoke_solver.rel",
    "batch_config/workflow/steps/load_data.rel",
    "batch_config/workflow/steps/materialize.rel"
]

COMMON_MODEL_RELATIVE_PATH = "/../rel"

CONFIG_BASE_RELATION = "batch:config"

DATE_FORMAT = "%Y%m%d"

DATE_PREFIX = "data_dt="  # TODO use pattern?

BLOB_PAGE_SIZE = 500

# RAI constants
IMPORT_CONFIG_REL = "import_config"

MISSED_RESOURCES_REL = "missing_resources_json"
WORKFLOW_JSON_REL = "workflow_json"
BATCH_CONFIG_REL = "batch:config"

FILE_LOAD_RELATION = {
    "CSV": "load_csv",
    "JSONL": "load_jsonlines_general",
    "JSON": "load_json" # TODO: maybe load_json_general?
}
# Env config params
RAI_PROFILE = "RAI_PROFILE"
RAI_PROFILE_PATH = "RAI_PROFILE_PATH"
RAI_SDK_HTTP_RETRIES = "RAI_SDK_HTTP_RETRIES"
AZURE_EXPORT_ACCOUNT = "AZURE_EXPORT_ACCOUNT"
AZURE_EXPORT_CONTAINER = "AZURE_EXPORT_CONTAINER"
AZURE_EXPORT_DATA_PATH = "AZURE_EXPORT_DATA_PATH"
AZURE_EXPORT_NUM_FILES = "AZURE_EXPORT_NUM_FILES"
AZURE_EXPORT_SAS = "AZURE_EXPORT_SAS"
AZURE_IMPORT_ACCOUNT = "AZURE_IMPORT_ACCOUNT"
AZURE_IMPORT_CONTAINER = "AZURE_IMPORT_CONTAINER"
AZURE_IMPORT_DATA_PATH = "AZURE_IMPORT_DATA_PATH"
AZURE_IMPORT_SAS = "AZURE_IMPORT_SAS"

# Step parameters
REL_CONFIG_DIR = "rel_config_dir"
OUTPUT_ROOT = "output_root"
LOCAL_DATA_DIR = "local_data_dir"
START_DATE = "start_date"
END_DATE = "end_date"
COLLAPSE_PARTITIONS_ON_LOAD = "collapse_partitions_on_load"



import glob
import logging
import os.path
import pathlib
from typing import List
from datetime import datetime

from workflow.common import EnvConfig
from workflow import blob, constants


class PathsBuilder:

    def build(self, logger: logging.Logger, date_range: List[datetime], relative_path: List[str], input_format,
              is_master: bool) -> List[str]:
        paths = self._build(logger, date_range, relative_path, input_format, is_master)
        if not paths:
            raise AssertionError(f"""PathsBuilder didn't find any file for specified parameters:
            date_range=`{date_range}`, relative_path=`{relative_path}`, 
            input_format=`{input_format}`, is_master=`{is_master}`""")
        return paths

    def _build(self, logger: logging.Logger, date_range: List[datetime], relative_path: List[str], input_format,
               is_master: bool) -> List[str]:
        raise NotImplementedError("This class is abstract")


class LocalPathsBuilder(PathsBuilder):
    local_data_dir: str

    def __init__(self, local_data_dir):
        self.local_data_dir = local_data_dir

    def _build(self, logger: logging.Logger, date_range: List[datetime], relative_path, extensions: List[str],
               is_master: bool) -> List[str]:
        paths = []
        files_path = f"{self.local_data_dir}/{relative_path}"
        if is_master:
            paths = self._get_folder_paths(files_path, extensions)
        else:
            for day in date_range:
                d = day.strftime(constants.DATE_FORMAT)
                folder_path = f"{files_path}/{constants.DATE_PREFIX}{d}"
                paths.extend(self._get_folder_paths(folder_path, extensions))

        paths = [os.path.abspath(path) for path in paths]
        return paths

    @staticmethod
    def _get_folder_paths(folder_path: str, extensions: List[str]):
        paths = []
        if pathlib.Path(folder_path).is_dir():
            for ext in extensions:
                paths.extend(glob.glob(f"{folder_path}/*.{ext}"))
        return paths


class RemotePathsBuilder(PathsBuilder):
    env_config: EnvConfig

    def __init__(self, env_config: EnvConfig):
        self.env_config = env_config

    def _build(self, logger: logging.Logger, date_range: List[datetime], relative_path, extensions: List[str],
               is_master: bool) -> List[str]:
        import_data_path = self.env_config.azure_import_data_path
        files_path = f"{import_data_path}/{relative_path}"

        logger.info(f"Loading data from blob import path: '{import_data_path}'")

        paths = []
        if is_master:
            paths = blob.list_files_in_containers(logger, self.env_config, files_path)
        else:
            for day in date_range:
                d = day.strftime(constants.DATE_FORMAT)
                logger.debug(f"Day from range: {d}")
                paths += blob.list_files_in_containers(logger, self.env_config,
                                                       f"{files_path}/{constants.DATE_PREFIX}{d}")
        return paths
